/*
  This file is part of ZitherLayout.

  ZitherLayout is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  ZitherLayout is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ZitherLayout.  If not, see <http://www.gnu.org/licenses/>.


  Diese Datei ist Teil von ZitherLayout.

  ZitherLayout ist Freie Software: Sie können es unter den Bedingungen
  der GNU General Public License, wie von der Free Software Foundation,
  Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren
  veröffentlichten Version, weiterverbreiten und/oder modifizieren.

  ZitherLayout wird in der Hoffnung, dass es nützlich sein wird, aber
  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
  Siehe die GNU General Public License für weitere Details.

  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
*/

/*
**  © 2011-2012 Carsten Gaede (PiperSoft)
**
**  @project: ZitherLayout
**  @file: ZitherLayout.c
**  @brief main module of ZitherLayout
**
**  @author: Carsten Gaede
**  @date: 2011-11-03
**
**
**  @version 1.0.0.0 2012-01-01 Carsten Gaede
**    first official release
**    added default settings file
**  @version 0.0.0.1 2011-11-03 Carsten Gaede
**    initial creation
*/

#define VERSION "1.0.0.0"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "input.h"
#include "layout.h"
#include "output.h"

#define DEFAULT_SETTINGS_FILENAME ".PiperSoft/ZitherLayout/default-settings"  // located in home directory

//*** dump routines ************************************************************

int DumpNote(struct NOTE *lpNote, int idx)
{
  printf("D: note[%03i]: len=%3i, val=%2i, acc=%2i, dir=%i, row=%2i, line_pos=%i, chord_pos=%i\r\n", idx, lpNote->length, lpNote->value, lpNote->chord, lpNote->dir, lpNote->pos.row, lpNote->line_pos_in, lpNote->chord_pos);
  return(0);
}

int DumpSong(struct SONG *song)
{
  int iNote;
  for(iNote = 0; iNote < song->notes.count; ++iNote) {
    DumpNote(&(song->notes.notes[iNote]), iNote);
  }
  printf("D: rows needed: %i\r\n", song->rows_needed);
  return(0);
}

#define DUMP(x) \
  printf("D: %s=%i\r\n", #x, s->set.x)
int DumpLayoutSettings(struct SESSION *s)
{
  DUMP(dim.page.height);
  DUMP(dim.page.width);
  DUMP(dim.page.borders.top);
  DUMP(dim.page.borders.bottom);
  DUMP(dim.page.borders.left);
  DUMP(dim.page.borders.right);
  DUMP(dim.picture.height);
  DUMP(dim.picture.width);
  DUMP(dim.picture.notes.width);
  DUMP(dim.picture.notes.offset);
  DUMP(dim.picture.notes.borders.top);
  DUMP(dim.picture.notes.borders.bottom);
  DUMP(dim.picture.notes.borders.left);
  DUMP(dim.picture.notes.borders.right);
  DUMP(dim.picture.info.width);
  DUMP(dim.picture.info.borders.top);
  DUMP(dim.picture.info.borders.bottom);
  DUMP(dim.picture.row.height);
  DUMP(dim.picture.col.width);
  DUMP(dim.notes.width);
  DUMP(dim.notes.width_dotted);
  DUMP(count.cols);
  DUMP(count.rows);

  return(0);
}

//*** main routine *************************************************************

int main(int argc, char **argv)
{
  int iResult = 0;

  struct SESSION session;
  FILE *fIn = (FILE*)NULL;
  FILE *fOut = stdout;
  int i;

  char *lpHome;
  char lpDefaultSettingsFileName[PATH_MAX];

  // clear session
  memset(&session, 0, sizeof(struct SESSION));

  //*** input ***

  // read command line arguments
  if (ReadCommandLineParameters(argc, argv, &session) < 0) {
    return(-1);
  }

  // print welcome message
  if (session.uLogMode >= LOG_MODE_NORMAL) {
    fprintf(stdout, "--- welcome to ZitherLayout by PiperSoft (version %s) ---\r\n", VERSION);
  }

  // set default values
  SetDefaultInputOptions(&session);
  SetDefaultLayoutOptions(&session);

  // try to read default settings file
  if ((lpHome = getenv("HOME")) != NULL) {
    sprintf(lpDefaultSettingsFileName, "%s/%s", lpHome, DEFAULT_SETTINGS_FILENAME);
    if ((fIn = fopen(lpDefaultSettingsFileName, "r")) != NULL) {
      if (session.uLogMode >= LOG_MODE_VERBOSE) {
        fprintf(stdout, "D: reading default settings from '%s'\r\n", lpDefaultSettingsFileName);
      }
      if (ReadInputFile(fIn, &session) < 0) {
        fprintf(stderr, "E: invalid default settings file '%s'\r\n", lpDefaultSettingsFileName);
        iResult = -1;
      }
      fclose(fIn);
      if (iResult < 0) {
        return(iResult);
      }
    }
  }

  // read input files
  for (i = 0; i < session.input.uFileCount; ++i) {
    if ((fIn = fopen(session.input.lpFileNames[i], "r")) == NULL) {
      fprintf(stderr, "E: could not open input file '%s'\r\n", session.input.lpFileNames[i]);
      return(-1);
    }
    if (session.uLogMode >= LOG_MODE_VERBOSE) {
      fprintf(stdout, "D: reading input file '%s'\r\n", session.input.lpFileNames[i]);
    }
    if (ReadInputFile(fIn, &session) < 0) {
      fprintf(stderr, "E: invalid input file '%s'\r\n", session.input.lpFileNames[i]);
      iResult = -1;
    }
    fclose(fIn);
    if (iResult < 0) {
      return(iResult);
    }
  }
  if (session.input.bReadStdIn) {
    if (ReadInputFile(stdin, &session) < 0) return(-1);
  }

  // apply settins from command line
  ApplyCommandLineSettings(&session);

  //*** layout ***
  GenerateLayout(&session);
  if (session.uLogMode >= LOG_MODE_NORMAL) {
    fprintf(stdout, "I: found %i notes\r\n", session.song.notes.count);
  }
  if (session.uLogMode >= LOG_MODE_VERBOSE) {
    DumpSong(&session.song);
    DumpLayoutSettings(&session);
  }

  //*** output ***

  // create output file
  if (session.output.lpFileName != NULL) {
    if ((fOut = fopen(session.output.lpFileName, "w")) == NULL) {
      fprintf(stderr, "E: could not create output file '%s'\r\n", session.output.lpFileName);
      return(-1);
    }
  }

  // export song
  ExportLaTeX(fOut, &session);

  // close output file
  if ((fOut != NULL) && (fOut != stdout)) {
    fclose(fOut);
  }

  // print bye bye message
  if (session.uLogMode >= LOG_MODE_NORMAL) {
    fprintf(stdout, "--- bye ---\r\n");
  }

	return EXIT_SUCCESS;
}
