/*
  This file is part of ZitherLayout.

  ZitherLayout is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  ZitherLayout is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ZitherLayout.  If not, see <http://www.gnu.org/licenses/>.


  Diese Datei ist Teil von ZitherLayout.

  ZitherLayout ist Freie Software: Sie können es unter den Bedingungen
  der GNU General Public License, wie von der Free Software Foundation,
  Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren
  veröffentlichten Version, weiterverbreiten und/oder modifizieren.

  ZitherLayout wird in der Hoffnung, dass es nützlich sein wird, aber
  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
  Siehe die GNU General Public License für weitere Details.

  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
*/

/*
**  © 2011-2012 Carsten Gaede (PiperSoft)
**
**  @project: ZitherLayout
**  @file: input.c
**  @brief implementation of input module
**
**  @date: 2011-11-13
**  @author: Carsten Gaede
**
**
**  @version 1.0.0.0 2012-01-01 Carsten Gaede
**    first official release
**  @version 0.0.0.1 2011-11-13 Carsten Gaede
**    initial creation
*/

#include "input.h"

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define INPUT_TYPE_NONE 0
#define INPUT_TYPE_CMD  1
#define INPUT_TYPE_INFO 2
#define INPUT_TYPE_NOTE 3


// definition of states
#define STATE_INIT    0
#define STATE_LENGTH  1
#define STATE_VALUE   2
#define STATE_OCTAVE  3
#define STATE_ACCORD  4
#define STATE_END     5
#define STATE_MASK_KEEP_POSITION  0x80

//                        a   b   c  d  e  f  g  h
unsigned char PITCHES[] = {10, 11, 1, 3, 5, 6, 8, 12};


//*** initialize default settings **********************************************

int SetDefaultInputOptions(struct SESSION *s)
{
  s->set.defaults.length = LENGTH_4;
  s->set.defaults.octave = 1;

  return(0);
}


//*** generic read methods *****************************************************

// generic read value makro
#define READVAL(p,v) \
  v = 0;\
  while ((*p >= '0') && (*p <= '9')) {\
    v *= 10;\
    v += (*p - '0');\
    ++p;\
  }

#define CHECK_LENGTH(x) \
  case x:\
    return(LENGTH_##x);\
  break;
unsigned char GetLength(int iLength)
{
  switch(iLength) {
    CHECK_LENGTH(1);
    CHECK_LENGTH(2);
    CHECK_LENGTH(4);
    CHECK_LENGTH(8);
    CHECK_LENGTH(16);
    CHECK_LENGTH(32);
    CHECK_LENGTH(64);
    //CHECK_LENGTH(128);
    //CHECK_LENGTH(256);
    default:
      return(LENGTH_INVALID);
    break;
  }
}


//*** read command line ********************************************************

struct option LONG_OPTIONS[] = {
  {"quiet", no_argument, 0, 'q'},
  {"verbose", no_argument, 0, 'v'},
  {"settings-file", required_argument, 0, 's'}, // settings file
  {"set", required_argument, 0, 'S'},           // settings from command line
  {"input-file", required_argument, 0, 'i'},    // input file, if none given, use stdin
  {"output-file", required_argument, 0, 'o'},   // output file, if none given, use stdout
  {0, 0, 0, 0}
};
#define OPTION_STRING "qvs:S:i:o:"

int ReadCommandLineParameters(int argc, char **argv, struct SESSION *s)
{
  int arg;

  s->input.bReadStdIn = 1;
  s->uLogMode = LOG_MODE_NORMAL;

  while ((arg = getopt_long(argc, argv, OPTION_STRING, LONG_OPTIONS, NULL)) >= 0) {
    switch(arg) {
      case '?':
        fprintf(stderr, "E: see README file\r\n");
        return(-1);
      break;
      case 'q':
        s->uLogMode = LOG_MODE_QUIET;
      break;
      case 'v':
        s->uLogMode = LOG_MODE_VERBOSE;
      break;
      case 'i':
        s->input.bReadStdIn = 0;
      case 's':
        if (s->input.uFileCount >= MAX_INPUT_FILES) {
          fprintf(stderr, "E: too many input files\r\n");
          return(-1);
        }
        s->input.lpFileNames[s->input.uFileCount] = optarg;
        ++s->input.uFileCount;
      break;
      case 'S':
        if (s->set.cmd_line.count >= MAX_CL_SETTINGS) {
          fprintf(stderr, "E: too many command line settings\r\n");
          return(-1);
        }
        s->set.cmd_line.items[s->set.cmd_line.count] = optarg;
        ++s->set.cmd_line.count;
      break;
      case 'o':
        if (s->output.lpFileName != NULL) {
          fprintf(stderr, "E: more than one output file given\r\n");
          return(-1);
        }
        s->output.lpFileName = optarg;
      break;
    }
  }

  return(0);
}


//*** read input files *********************************************************

// reserved: end, defined: dl, do, pb, pb[tblr], nb[lr], lp, ln, g
/*
 *  dl  : default length
 *  do  : default octave
 *  pb  : page borders
 *  pbt : page border top
 *  pbb :             bottom
 *  pbl :             left
 *  pbr :             right
 *  nbl : note border left
 *  nbr :             right
 *  lp  : lyrics position
 *  ln  : lyric numbers
 *  g   : grid
 *  sd  : string distance
 *  sc  : string count (= pitch count)
 *  lc  : lowest compression
 *  hc  : highest compression
 *  fc  : fix compression (sets lowest and highest)
 *  cc  : chord count
 *  np  : notes position  ((t)op / (c)enter / (b)ottom)
 *
 *  cew : cut width
 *  ceh : cut height
*/
int ReadCommand(char *lpInput, struct SESSION *s)
{
  char *pval;
  unsigned short value = 0;

  //*** read style ***
  if (strncmp(lpInput, "s=", 2) == 0) {
    pval = lpInput + 2;
    switch(*pval) {
      case 'm':
        s->set.layout_style = LAYOUT_STYLE_MELODY;
      break;
      case 'b':
        s->set.layout_style = LAYOUT_STYLE_BASS;
      break;
      default:
        fprintf(stderr, "W: invalid layout style, input: %s\r\n", lpInput);
      break;
    }
  }
  //*** input defaults ***
  else if (strncmp(lpInput, "dl=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    if ((s->set.defaults.length = GetLength(value)) == LENGTH_INVALID) {
     fprintf(stderr, "W: invalid default length, input: %s\r\n", lpInput);
    }
  }
  else if (strncmp(lpInput, "do=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    if ((value >= 1) && (value <= 3)) {
      s->set.defaults.octave = value;
    }
    else {
      fprintf(stderr, "W: invalid default octave, input: %s\r\n", lpInput);
    }
  }
  /*else if (strncmp(lpLine, "dos=", 4) == 0) {

  }*/
  //*** dimensions ***
  else if (strncmp(lpInput, "pb=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.dim.page.borders.top = value;
    s->set.dim.page.borders.bottom = value;
    s->set.dim.page.borders.left = value;
    s->set.dim.page.borders.right = value;
  }
  else if (strncmp(lpInput, "pbt=", 3) == 0) {
    pval = lpInput + 4;
    READVAL(pval, value);
    s->set.dim.page.borders.top = value;
  }
  else if (strncmp(lpInput, "pbb=", 3) == 0) {
    pval = lpInput + 4;
    READVAL(pval, value);
    s->set.dim.page.borders.bottom = value;
  }
  else if (strncmp(lpInput, "pbl=", 3) == 0) {
    pval = lpInput + 4;
    READVAL(pval, value);
    s->set.dim.page.borders.left = value;
  }
  else if (strncmp(lpInput, "pbr=", 3) == 0) {
    pval = lpInput + 4;
    READVAL(pval, value);
    s->set.dim.page.borders.right = value;
  }
  else if (strncmp(lpInput, "nbl=", 3) == 0) {
    pval = lpInput + 4;
    READVAL(pval, value);
    s->set.dim.picture.notes.borders.left = value;
  }
  else if (strncmp(lpInput, "nbr=", 3) == 0) {
    pval = lpInput + 4;
    READVAL(pval, value);
    s->set.dim.picture.notes.borders.right = value;   // melody style
    s->set.dim.picture.info.c1_dist = value;          // bass style
  }
  else if (strncmp(lpInput, "sd=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.dim.picture.col.width = value;             // melody style
    s->set.dim.chords.dist = value;                  // bass style
  }
  else if (strncmp(lpInput, "cw=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.dim.chords.width = value;
  }
  else if (strncmp(lpInput, "cd=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.dim.chords.c1_dist = value;
  }
  //*** layout options ***
  else if (strncmp(lpInput, "sc=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.count.cols = value;
  }
  else if (strncmp(lpInput, "cc=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.count.chords = value;
  }
  else if (strncmp(lpInput, "np=", 3) == 0) {
    pval = lpInput + 3;
    switch(*pval) {
      case 't':
        s->set.notes_position = NOTES_POS_TOP;
      break;
      case 'c':
        s->set.notes_position = NOTES_POS_CENTER;
      break;
      case 'b':
        s->set.notes_position = NOTES_POS_BOTTOM;
      break;
      default:
        fprintf(stderr, "W: invalid notes position, input %s\r\n", lpInput);
      break;
    }
  }
  else if (strncmp(lpInput, "lc=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.compression.min = value;
  }
  else if (strncmp(lpInput, "hc=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.compression.max = value;
  }
  else if (strncmp(lpInput, "fc=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.compression.min = value;
    s->set.compression.max = value;
  }
  //*** lyrics ***
  else if (strncmp(lpInput, "lp", 2) == 0) {
    pval = lpInput + 2;
    READVAL(pval, value);
    if ((*pval == '=') && (value >= 1) && (value <= MAX_LYRICS)) {
      ++pval;
      READVAL(pval, s->song.lyrics.lyrics[value-1].pos.col);
      if (*pval == ',') {
        ++pval;
        READVAL(pval, s->song.lyrics.lyrics[value-1].pos.row);
      }
    }
    else {
      fprintf(stderr, "W: invalid lyrics number, input %s\r\n", lpInput);
    }
  }
  else if (strncmp(lpInput, "ln=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.lyric_numbers = value & 0xFF;
  }
  //*** grid ***
  else if (strncmp(lpInput, "g=", 2) == 0) {
    pval = lpInput + 2;
    READVAL(pval, value);
    s->set.grid = value & 0xFF;
  }
  else if (strncmp(lpInput, "cew=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.dim.cut.width = value;
  }
  else if (strncmp(lpInput, "ceh=", 3) == 0) {
    pval = lpInput + 3;
    READVAL(pval, value);
    s->set.dim.cut.height = value;
  }
  else {
    fprintf(stderr, "W: ignoring unknown command, input: %s\r\n", lpInput);
  }

  return(0);
}

int ReadSongInfo(char *lpLine, struct SONG *lpSong)
{
  int len = strlen(lpLine);
  char *pos;
  int value;
  if (strncmp(lpLine, "t=", 2) == 0) {
    len -= 2;
    strncpy(lpSong->title, lpLine + 2, MIN(len, MAXLEN_TITLE));
  }
  else if (strncmp(lpLine, "s=", 2) == 0) {
    len -= 2;
    strncpy(lpSong->subtitle, lpLine + 2, MIN(len, MAXLEN_SUBTITLE));
  }
  else if (strncmp(lpLine, "a=", 2) == 0) {
    len -= 2;
    strncpy(lpSong->artists.composer, lpLine + 2, MIN(len, MAXLEN_ARTIST));
    strncpy(lpSong->artists.texter, lpLine + 2, MIN(len, MAXLEN_ARTIST));
  }
  else if (strncmp(lpLine, "c=", 2) == 0) {
    len -= 2;
    strncpy(lpSong->artists.composer, lpLine + 2, MIN(len, MAXLEN_ARTIST));
  }
  else if (strncmp(lpLine, "te=", 2) == 0) {
    len -= 3;
    strncpy(lpSong->artists.texter, lpLine + 3, MIN(len, MAXLEN_ARTIST));
  }
  else if (strncmp(lpLine, "arr=", 2) == 0) {
    len -= 4;
    strncpy(lpSong->artists.arrangement, lpLine + 4, MIN(len, MAXLEN_ARRANGEMENT));
  }
  else if (strncmp(lpLine, "w=", 2) == 0) {
    len -= 2;
    strncpy(lpSong->warning, lpLine + 2, MIN(len, MAXLEN_WARNING));
  }
  else if (strncmp(lpLine, "h=", 2) == 0) {
    len -= 2;
    strncpy(lpSong->hint, lpLine + 2, MIN(len, MAXLEN_HINT));
  }
  else if (strncmp(lpLine, "l", 1) == 0) {
    pos = lpLine + 1;
    READVAL(pos, value);
    if ((*pos == '=') && (value >= 1) && (value <= MAX_LYRICS)) {
      ++pos;
      len = strlen(pos);
      strncpy(lpSong->lyrics.lyrics[value - 1].text, pos, MIN(len, MAXLEN_LYRICS));
      lpSong->lyrics.count = MAX(lpSong->lyrics.count, value);
    }
    else {
      fprintf(stderr, "W: invalid lyrics number, input: %s\r\n", lpLine);
    }
  }
  else {
    fprintf(stderr, "W: ignoring unknown info, input: %s\r\n", lpLine);
  }

  return(0);
}

int ReadNote(char *lpLine, struct SETTINGS *set, struct NOTE *lpNote)
{
  int iState = 0, len = 0, oct = 0, chord = 0;
  char *lpPos = lpLine;
  struct INPUT_DEFAULTS *def = &set->defaults;

  iState = STATE_INIT;
  while ((*lpPos != '\0') && (*lpPos != ' ') && (*lpPos != '\t') && (*lpPos != ';')) {
    switch(iState) {
      case STATE_INIT:
        lpNote->length = LENGTH_NONE;
        lpNote->value = 0;
        iState = STATE_LENGTH | STATE_MASK_KEEP_POSITION;
        len = oct = chord = 0;
      break;
      case STATE_LENGTH:
        if ((*lpPos >= '0') && (*lpPos <= '9')) {
          len *= 10;
          len += (*lpPos - '0');
        }
        else if (*lpPos == '.') {
          lpNote->length |= LENGTH_PT;
        }
        /*else if (*lpPos == ':') {
          lpNote->length |= LENGTH_DPT;
        }*/
        else {
          iState = STATE_VALUE | STATE_MASK_KEEP_POSITION;
        }
      break;
      case STATE_VALUE:
        if (*lpPos == '-') {
          lpNote->value -= 1;
        }
        else if (*lpPos == '+') {
          lpNote->value += 1;
        }
        else if ((*lpPos >= 'a') && (*lpPos <= 'h')) {
          lpNote->value += PITCHES[*lpPos - 'a'];
          iState = STATE_OCTAVE;
        }
        else if (*lpPos == 'p') { // pause
          lpNote->value = VALUE_PAUSE;
          iState = STATE_END;
        }
        else {
          iState = STATE_OCTAVE | STATE_MASK_KEEP_POSITION;
        }
      break;
      case STATE_OCTAVE:
        if ((*lpPos >= '0') && (*lpPos <= '9')) {
          oct *= 10;
          oct += (*lpPos - '0');
        }
        else if ((*lpPos == 'A') || (*lpPos == 'C')) {
          iState = STATE_ACCORD;
        }
        else {
          iState = STATE_END | STATE_MASK_KEEP_POSITION;
        }
      break;
      case STATE_ACCORD:
        if ((*lpPos >= '0') && (*lpPos <= '9')) {
          chord *= 10;
          chord += (*lpPos - '0');
        }
        else if (*lpPos == '.') {
          lpNote->chord |= CHORD_PT;
        }
        else if (*lpPos == '_') {
          lpNote->chord |= CHORD_SLR;
        }
        else {
          iState = STATE_END | STATE_MASK_KEEP_POSITION;
        }
      break;
    }
    if (!(iState & STATE_MASK_KEEP_POSITION)) {
      ++lpPos;
    }
    iState &= ~STATE_MASK_KEEP_POSITION;
  }
  // apply length
  if (len == 0) {
    if (def->length == 0) {
      fprintf(stderr, "E: no length given, input: %s\r\n", lpLine);
      return(-1);
    }
    //printf("D: using default length (%i), input: %s\r\n", set->def.len, lpLine);
    lpNote->length |= def->length;
  }
  else {
    if ((len = GetLength(len)) == LENGTH_INVALID) {
      fprintf(stderr, "E: invalid length, input: %s\r\n", lpLine);
      return(-1);
    }
    lpNote->length |= len;
  }
  // apply octave
  if (oct == 0) {
    if ((def->octave == 0)/* && (def->octave_start == 0)*/) {
      fprintf(stderr, "E: no octave given, input: %s\r\n", lpLine);
      return(-1);
    }
    if (def->octave != 0) {
      //printf("D: using default octave (%i), input: %s\r\n", set->def.oct, lpLine);
      oct = def->octave;
    }
  }
  lpNote->value += (12 * (oct - 1));
  // apply accord
  if (chord != 0) {
    if ((chord >= 1) && (chord <= set->count.chords)) {
      lpNote->chord |= chord;
    }
    else {
      fprintf(stderr, "W: ignoring invalid accord, input: %s\r\n", lpLine);
    }
  }
  //printf("D: input line '%s', len=%i, val=%i\r\n", lpLine, lpNote->len, lpNote->value);

  return(0);
}


//*** reading input file ***

int ReadInputFile(FILE *lpFile, struct SESSION *s)
{
  int res;
  char lpLine[MAX_INPUT_LENGTH];
  struct NOTES *lpNotes;
  char *pos;
  unsigned char itype;

  lpNotes = &s->song.notes;
  memset(lpLine, '\0', MAX_INPUT_LENGTH);
  while (fgets(lpLine, MAX_INPUT_LENGTH, lpFile) != NULL) {
    while ((lpLine[strlen(lpLine)-1] == '\r') || (lpLine[strlen(lpLine)-1] == '\n')) {
      lpLine[strlen(lpLine)-1] = '\0';
    }
    //printf("D: input line '%s'\r\n", line);
    pos = lpLine;
    while ((*pos == ' ') || (*pos == '\t')) {
      ++pos;
    }
    if (*pos == '\0') continue; // ignore empty line
    if (*pos == ';') continue;  // ignore comment line
    if (strcmp(lpLine, "\\end") == 0) return(0);
    switch(*pos) {
      case '\\':
        itype = INPUT_TYPE_CMD;
        ++pos;
      break;
      case '@':
        itype = INPUT_TYPE_INFO;
        ++pos;
      break;
      default:
        itype = INPUT_TYPE_NOTE;
      break;
    }
    while (*pos != '\0') {
      while ((*pos == ' ') || (*pos == '\t')) ++pos;  // skip leading white space
      if (*pos == '\0') break;  // break at end of line
      if (*pos == ';') break;   // break line at comment
      switch(itype) {
        case INPUT_TYPE_CMD:
          if ((res = ReadCommand(pos, s)) < 0) {
            return(res);
          }
          else if (res == 1) {
            return(0);
          }
        break;
        case INPUT_TYPE_INFO:
          if (ReadSongInfo(pos, &s->song) != 0) {
            return(-1);
          }
        break;
        case INPUT_TYPE_NOTE:
          if (lpNotes->count == MAX_NOTES) {
            fprintf(stderr, "W: maximum number of notes reached, input: %s\r\n", pos);
            return(0);
          }
          if (ReadNote(pos, &s->set, &lpNotes->notes[lpNotes->count]) != 0) {
            return(-1);
          }
          ++lpNotes->count;
        break;
      }
      if (itype == INPUT_TYPE_INFO) break;  // info ends at line end
      while ((*pos != ' ') && (*pos != '\t') && (*pos != ';') && (*pos != '\0')) ++pos;  // go to next item
    }
    memset(lpLine, '\0', MAX_INPUT_LENGTH);
  }

  return(0);
}


//*** apply command line settings **********************************************

int ApplyCommandLineSettings(struct SESSION *s)
{
  int i;
  for (i = 0; i < s->set.cmd_line.count; ++i) {
    ReadCommand(s->set.cmd_line.items[i], s);
  }

  return(0);
}
