/*
  This file is part of ZitherLayout.

  ZitherLayout is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  ZitherLayout is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ZitherLayout.  If not, see <http://www.gnu.org/licenses/>.


  Diese Datei ist Teil von ZitherLayout.

  ZitherLayout ist Freie Software: Sie können es unter den Bedingungen
  der GNU General Public License, wie von der Free Software Foundation,
  Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren
  veröffentlichten Version, weiterverbreiten und/oder modifizieren.

  ZitherLayout wird in der Hoffnung, dass es nützlich sein wird, aber
  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
  Siehe die GNU General Public License für weitere Details.

  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
*/

/*
**  © 2011-2012 Carsten Gaede (PiperSoft)
**
**  @project: ZitherLayout
**  @file: layout.c
**  @brief implementation of layout module
**
**  @date: 2011-11-14
**  @author: Carsten Gaede
**
**
**  @version 1.0.0.0 2012-01-01 Carsten Gaede
**    first official release
**  @version 0.0.0.1 2011-11-14 Carsten Gaede
**    initial creation
*/

#include "layout.h"

#include <stdio.h>


// definition of directions
#define DIR_NONE  0x00
#define DIR_STAY  0x01
#define DIR_UP    0x02
#define DIR_DOWN  0x03
#define DIR_TURN  0x10
#define DIR_UTURN (DIR_UP | DIR_TURN)
#define DIR_DTURN (DIR_DOWN | DIR_TURN)

// list of chord positions (0: invalid, impossible)
//   first index is line_pos_in (lp)
//   second index is line_pos_in of next note (nlp)
unsigned char ACCRPOS[6][6] = {
// lp \nlp: 0,      L,      TL,     T,      TR,     R
//      to: 0,      R,      BR,     B,      BL,     L
  /*0*/   { RPOS_L, RPOS_L, RPOS_L, RPOS_L, RPOS_R, RPOS_R},
  /*L*/   { RPOS_R, RPOS_B, RPOS_B, RPOS_R, 0,      0},
  /*TL*/  { RPOS_R, RPOS_B, RPOS_B, RPOS_R, RPOS_R, 0},
  /*T*/   { RPOS_L, RPOS_L, RPOS_L, RPOS_L, RPOS_R, RPOS_R},
  /*TR*/  { RPOS_L, 0,      RPOS_L, RPOS_L, RPOS_B, RPOS_B},
  /*R*/   { RPOS_L, 0,      0,      RPOS_L, RPOS_B, RPOS_B}
};


//*** setting default dimensions ***********************************************

/*            norm  smal  tiny
note height   7010  5608  4474
note width    2204  1763  1406
note w. dotd  3439  2751  2195
row height    4000  3000  2500
*/
int SetDimensions(struct SESSION *s, unsigned char compression)
{
  /*  try to compress by 0.1mm row height, from max row height to 30 use normal notes
   * from 30 to 26 use small notes
   * from 25 to xx use tiny notes
   * */
  s->set.dim.picture.row.height = DIM_DEF_MAX_ROW_HEIGHT - compression;
  if (compression <= MAX_COMPRESSION_NORMAL_SIZE) {
    s->set.dim.notes.width        = 22;
    s->set.dim.notes.width_dotted = 35;
    s->set.dim.notes.chord.width    = 20;
    s->set.dim.notes.chord.height   = 30;
  }
  else if (compression <= MAX_COMPRESSION_SMALL_SIZE) {
    s->set.dim.notes.width        = 18;
    s->set.dim.notes.width_dotted = 28;
    s->set.dim.notes.chord.width    = 16;
    s->set.dim.notes.chord.height   = 25;
  }
  else {
    s->set.dim.notes.width        = 14;
    s->set.dim.notes.width_dotted = 22;
    s->set.dim.notes.chord.width    = 12;
    s->set.dim.notes.chord.height   = 20;
  }
  s->set.dim.picture.notes.borders.top = s->set.dim.picture.row.height / 2;
  s->set.dim.picture.notes.borders.bottom = s->set.dim.picture.row.height / 2;

  return(0);
}

int SetDefaultLayoutOptions(struct SESSION *s)
{
  // default page dimensions and borders
  s->set.dim.page.height          = DIM_DEF_PAGE_HEIGHT;
  s->set.dim.page.width           = DIM_DEF_PAGE_WIDTH;
  s->set.dim.page.borders.top     = DIM_DEF_PAGE_BORDER;
  s->set.dim.page.borders.bottom  = DIM_DEF_PAGE_BORDER;
  s->set.dim.page.borders.left    = DIM_DEF_PAGE_BORDER;
  s->set.dim.page.borders.right   = DIM_DEF_PAGE_BORDER;
  // default note borders, row and column sizes
  s->set.dim.picture.notes.borders.right  = DIM_DEF_BORDER_RIGHT;
  s->set.dim.picture.col.width = DIM_DEF_COL_WIDTH;

  s->set.dim.picture.info.c1_dist = DIM_DEF_INFO_C1_DIST;
  s->set.dim.picture.info.title.border = DIM_DEF_TITLE_BORDER;
  s->set.dim.picture.info.title.sub_title_dist = DIM_DEF_SUB_TITLE_DIST;

  s->set.dim.chords.width = DIM_DEF_ACCORD_WIDTH;
  s->set.dim.chords.c1_dist = DIM_DEF_ACCORD_C1_DIST;
  s->set.dim.chords.dist = DIM_DEF_ACCORD_DIST;

  s->set.dim.notes.dy = -9;
  s->set.dim.notes.chord.dist = DIM_DEF_ACC_DIST;
  s->set.dim.lyrics.number_dist = DIM_DEF_LYRIC_NUMBER_DIST;

  s->set.dim.cut.width = DIM_DEF_CUT_WIDTH;
  s->set.dim.cut.height = DIM_DEF_CUT_HEIGHT;

  s->set.count.cols = DEF_MAX_PITCH;
  s->set.count.chords = DEF_MAX_CHORD;

  s->set.notes_position = NOTES_POS_CENTER;
  s->set.compression.min = DEF_MIN_COMPRESSION;
  s->set.compression.max = DEF_MAX_COMPRESSION;

  return(0);
}


//*** calculate logical and relative positions *********************************

int CalculateDirections(struct SONG *song)
{
  struct NOTES *lpNotes = &song->notes;
  int iNote;
  unsigned char val, prev_val, next_val;
  unsigned char *dir;

  for (iNote = 0; iNote < lpNotes->count; ++iNote) {
    val = lpNotes->notes[iNote].value;
    prev_val = (iNote > 0) ? lpNotes->notes[iNote - 1].value : VALUE_NONE;
    next_val = (iNote < (lpNotes->count - 1)) ? lpNotes->notes[iNote + 1].value : VALUE_NONE;
    dir = &lpNotes->notes[iNote].dir;
    if (prev_val == VALUE_NONE) {
      // first note doesn't have direction
      *dir = DIR_STAY;
    }
    else if (prev_val == val) {
      // same value again
      *dir = DIR_STAY;
    }
    else if (prev_val < val) {
      if ((next_val == VALUE_NONE) || (val <= next_val)) {
        // going up
        *dir = DIR_UP;
      }
      else {
        // turn around, come from lower, going back to lower
        *dir = DIR_UP | DIR_TURN;
      }
    }
    else {  // prev_val > val
      if ((next_val == VALUE_NONE) || (val >= next_val)) {
        // going down
        *dir = DIR_DOWN;
      }
      else {
        // turn around, come from higher, going back to higher
        *dir = DIR_DOWN | DIR_TURN;
      }
    }
  }

  return(0);
}

int CalculateRelativeAccordPositions(struct SONG *song)
{
  int iNote;
  struct NOTES *lpNotes = &song->notes;
  struct NOTE *lpNote;
  unsigned char lpos, next_lpos;

  for (iNote = 0; iNote < lpNotes->count; ++iNote) {
    lpNote = &lpNotes->notes[iNote];
    if (lpNote->chord == 0) continue;
    lpos = lpNote->line_pos_in;
    next_lpos = RPOS_NONE;
    if (iNote < (lpNotes->count - 1)) {
      next_lpos = lpNotes->notes[iNote + 1].line_pos_in;
    }
    lpNote->chord_pos = ACCRPOS[lpos][next_lpos];
  }

  return(0);
}

int CalculateNotePositions(struct SONG *song)
{
  struct NOTES *lpNotes = &song->notes;
  int iNote;
  struct NOTE *lpNote;
  unsigned char prev_dir;
  unsigned short last_row;

  last_row = 0;
  prev_dir = DIR_NONE;
  for (iNote = 0; iNote < lpNotes->count; ++iNote) {
    lpNote = &lpNotes->notes[iNote];
    lpNote->pos.col = lpNote->value > 0 ? lpNote->value - 1 : 0;
    if (prev_dir == DIR_NONE) {   // beginning of the song
      lpNote->pos.row = 2;    // leave space for start needle
      lpNote->line_pos_in = RPOS_NONE;
    }
    else if (lpNote->dir == DIR_STAY) {
      lpNote->pos.row = last_row + 2;
      lpNote->line_pos_in = RPOS_TOP;
    }
    else if ((lpNote->dir & DIR_TURN) || (prev_dir & DIR_TURN)) {
      lpNote->pos.row = last_row + 1;
      if ((lpNote->dir == DIR_UTURN) || (prev_dir == DIR_DTURN)) {
        lpNote->line_pos_in = RPOS_TL;
      }
      else {
        lpNote->line_pos_in = RPOS_TR;
      }
    }
    else {
      lpNote->pos.row = last_row;
      if (lpNote->dir == DIR_UP) {
        lpNote->line_pos_in = RPOS_LEFT;
      }
      else {
        lpNote->line_pos_in = RPOS_RIGHT;
      }
    }
    last_row = lpNote->pos.row;
    prev_dir = lpNote->dir;
  }
  song->rows_needed = last_row + 2;   // leave space for last chord

  return(0);
}


//*** calculate layout parameters **********************************************

int CalculateFixLayoutParameters(struct SESSION *s)
{
  struct DIMENSIONS *dim = &s->set.dim;
  dim->picture.height = dim->page.height - dim->page.borders.top - dim->page.borders.bottom;
  dim->picture.width = dim->page.width - dim->page.borders.left - dim->page.borders.right;
  if (dim->picture.notes.borders.left == 0) {   // protect manual settings
    dim->picture.notes.borders.left = dim->picture.col.width / 2;
  }
  dim->picture.notes.width = dim->picture.notes.borders.left + dim->picture.notes.borders.right;
  dim->picture.notes.width += (s->set.count.cols - 1) * dim->picture.col.width;

  dim->picture.info.width = dim->picture.width - dim->picture.notes.width;
  dim->picture.notes.offset = dim->picture.info.width;

  return(0);
}

int CalculateDynamicLayoutParameters(struct SESSION *s)
{
  int i, j;
  for (i = s->set.compression.min; i <= s->set.compression.max; ++i) {
    s->set.compression.current = i;
    SetDimensions(s, i);
    s->set.count.rows = (s->set.dim.picture.height / s->set.dim.picture.row.height) + 1 - 1;  // one for top / bottom border
    if (s->song.rows_needed <= s->set.count.rows) {
      // notes fit on the page
      break;
    }
    else if (i < s->set.compression.max) {
      if (s->uLogMode >= LOG_MODE_NORMAL) {
        printf("I: switching to compression %i\r\n", i + 1);
      }
    }
  }
  if (s->set.count.rows < s->song.rows_needed) {
    fprintf(stderr, "W: not all notes will fit on the page\r\n");
    return(1);
  }
  else if (s->song.rows_needed < s->set.count.rows) { // center notes
    i = 0;
    switch(s->set.notes_position) {
      case NOTES_POS_TOP:
      break;
      case NOTES_POS_CENTER:
        i = (s->set.count.rows - s->song.rows_needed) / 2;
      break;
      case NOTES_POS_BOTTOM:
        i = s->set.count.rows - s->song.rows_needed;
      break;
    }
    if (i) {
      if (s->uLogMode >= LOG_MODE_VERBOSE) {
        printf("D: positioning notes, skipping %i rows\r\n", i);
      }
      for (j = 0; j < s->song.notes.count; ++j) {
        s->song.notes.notes[j].pos.row += i;
      }
    }
  }

  return(0);
}

int CalculateDynamicDimensions(struct SESSION *s)
{
  s->set.dim.picture.info.borders.top = s->set.dim.picture.notes.borders.top;
  s->set.dim.picture.info.borders.bottom = s->set.dim.picture.notes.borders.bottom;
  s->set.dim.notes.dx = -(s->set.dim.notes.width / 2);
  s->set.dim.notes.start_needle_size = s->set.dim.notes.width / 2;
  s->set.dim.notes.circle_size = (5 * s->set.dim.notes.width) / 2;
  s->set.dim.picture.info.title.width = s->set.dim.picture.height - s->set.dim.picture.notes.borders.top - s->set.dim.picture.notes.borders.bottom - (2 * s->set.dim.picture.info.title.border);

  s->set.dim.notes.chord.dx_l = -(s->set.dim.notes.chord.width + s->set.dim.notes.chord.dist) + s->set.dim.notes.dx;
  s->set.dim.notes.chord.dx_r = s->set.dim.notes.chord.dist + s->set.dim.notes.width_dotted + s->set.dim.notes.dx;
  s->set.dim.notes.chord.dy_lr = -(s->set.dim.notes.chord.height / 2);
  s->set.dim.notes.chord.dx_tb = -(s->set.dim.notes.chord.width / 2);
  s->set.dim.notes.chord.dy_b = -(s->set.dim.notes.chord.height + s->set.dim.notes.chord.dist) + s->set.dim.notes.dy;

  return(0);
}


//*** calculate real positions *************************************************

int CalculateRealPositions(struct SESSION *s)
{
  struct NOTES *lpNotes = &s->song.notes;
  int i;
  struct POSITION *pos;
  // calculate position from line and value
  for (i = 0; i < lpNotes->count; ++i) {
    pos = &s->song.notes.notes[i].pos;
    pos->x = s->set.dim.picture.notes.offset + s->set.dim.picture.notes.borders.left + (pos->col * s->set.dim.picture.col.width);
    pos->y = s->set.dim.picture.height - s->set.dim.picture.notes.borders.top - (pos->row * s->set.dim.picture.row.height);
  }
  if (lpNotes->count > 0) {
    s->song.start_needle_pos.x = s->song.notes.notes[0].pos.x;
    s->song.start_needle_pos.y = s->song.notes.notes[0].pos.y + (2 * s->set.dim.picture.row.height);
  }
  for (i = 0; i < s->song.lyrics.count; ++i) {
    pos = &s->song.lyrics.lyrics[i].pos;
    pos->x = s->set.dim.picture.notes.offset + s->set.dim.picture.notes.borders.left + (pos->col * s->set.dim.picture.col.width);
    pos->y = s->set.dim.picture.height - s->set.dim.picture.notes.borders.top - (pos->row * s->set.dim.picture.row.height);
  }

  return(0);
}


//*** generate complete layout *************************************************

int GenerateLayout(struct SESSION *s)
{
  CalculateDirections(&s->song);
  CalculateNotePositions(&s->song);
  CalculateRelativeAccordPositions(&s->song);

  CalculateFixLayoutParameters(s);
  CalculateDynamicLayoutParameters(s);
  CalculateDynamicDimensions(s);

  CalculateRealPositions(s);

  return(0);
}
